#include "bloomconfig.h"

#include <sys/types.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>

#include <openssl/sha.h>
#include <openssl/evp.h>
#include <openssl/md5.h>

#include "bloom.h"

#include <getopt.h>
#include <ctype.h>

const char *progname = "regression";
int opt_x = 0;
int opt_just = 0;

const char *hex2string(char *buf,int buflen,unsigned char *md,int len)
{
    const char *bufstart = buf;
    while(len>0 && buflen+1>0){
	sprintf(buf,"%02x",md[0]);
	md ++;
	buf +=2;
	buflen -= 2;
	len --;
    }
    return bufstart;
}

void string_increment(char *buf){
    int length = strlen(buf);
    char *walker = buf+length-1;  //start at end of string
    while(true){	
	if(*walker == 'z'){
	    *walker = '0';
	    walker --;
	}
	else{
	    *walker = (*walker)++;//increment last character
	    return;
	}	
	if(walker == (buf - sizeof(char))){walker = &buf[length-1];return;} //return from left end to right end
    }
}

void usage(){
    printf("Usage: %s [options] {bloomfile|:SEED:} \n", progname);
    printf(" Specify :S:M:k to create an in-memory bloom filter and pre-seed it with the input\n");
    printf(" Specify :E:M:k to create and use an in-memory bloom filter\n");
    printf("         m = 2^M\n");
    printf("     -q         Quiet output.\n");
    printf("     -b count   Bruteforce from seed looking for a match in filter. 0 for no count.\n");
    printf("     -x         don't look anything up in the bloom filter.\n");
    printf("     -i         Displays bloomfile information.\n");
    printf("     -j         Just bring the bloom filter into memory, then quit\n");
    printf("\n input selection:\n");
    printf("     -s string  Seed string.  Default is empty string. ("")\n");
    printf("     -n int     Number of hashes to generate and query.\n");
    printf("     -f filename Filename to output generated hashed to.\n");
    printf("\n");
    printf("     -h         Shows this usage.\n");
    exit(0);
}

/* Genprime implements a simple prime-number based random number generator */
void genprime(int count,unsigned char buf[20])
{
    /* Five big primes */
    u_int32_t prime[5] = {192991619, 192991627, 192991651, 192991663, 192991679};
    u_int32_t *num = (u_int32_t *)buf;
    for(u_int32_t i=0;i<5;i++){
	num[i] = count * prime[i];
    }
}

int main(int argc,char **argv)
{
    char message[1024];
    strncpy(message, "", 1);
    int ch=0; //for switch
    int info = 0;  //to print info
    int count = 1000000;  //how many iterations
    int	found = 0; 
    int quiet = 0;
    int brute = 0;
    int brutecount = 0;
 	unsigned char sha1[20];
    char *out_filename = 0;
    FILE *outfile = 0;
    const char *bloomfile = 0;//where bloom gets loaded from

    while((ch = getopt(argc, argv, "t:f:qs:mn:b:ihpxj")) != -1){
        switch(ch){
	case 'i': info = 1; break;
	case 'b': brute = 1; brutecount = atoi(optarg); break;
	case 'n': count = atoi(optarg); break;
	case 's': strncpy(message,optarg,40); found = 0; break;
	case 'q': quiet = 1; break;
	case 'f': out_filename = optarg; break;
	case 'x': opt_x = 1;break;
	case 'j': opt_just = 1;break;
	default:
	case 'h': usage(); 
        }//end switch
    }//end while loop for processing options

    argc -= optind;  //account for arguments handled
    argv += optind;  //walk to end of args list

    if(argc<1 && !out_filename) usage();
    if(out_filename){
	outfile = fopen(out_filename,"w");
	if(!outfile) {
	    exit(1);
	}
    }

    bloomfile = *argv++;
    argc--;
    
    nsrl_bloom *b = nsrl_bloom_alloc();
    if(bloomfile[0]==':'){
	char command;
	int M;
	int k;
	sscanf(bloomfile,":%c:%d:%d",&command,&M,&k);
	if(command!='S' && command!='E') errx(1,"Invalid bloom filter creation spec.");
	if(nsrl_bloom_create(b,0,160,M,k,"in-memory")){
	    errx(1,"Cannot create bloom filter with M=%d and k=%d",M,k);
	}
	if(command=='S'){
	    /* Seed the bloom filter */
	    for(int i=0; i<count; i++){
		unsigned char hashbuf[20];
		genprime(i,hashbuf);
		nsrl_bloom_add(b, hashbuf);
	    }
	}
    }
    else{
	if(access(bloomfile,R_OK)){
	    perror(bloomfile);
	    fprintf(stderr,"Invalid Bloom file.\n");
	    exit(1);
	}
	if(nsrl_bloom_open(b,bloomfile,O_RDONLY)){
	    err(1,"nsrl_bloom_open(%s)",bloomfile);
	}
    }

    if(opt_just){
      unsigned int counts[256];
      nsrl_calc_histogram(b,counts);
      printf("%s loaded into RAM\n",bloomfile);
      exit(0);
    }
    
    if(info){
        char ibuf[4096];
        nsrl_bloom_info(ibuf, sizeof(ibuf), b);
        fputs(ibuf, stdout);
    }

    /* The real regression loop */
    if(brute){
    	char nmessage[40];
    	for(int i=0; i<brutecount; i++){
    		SHA1((u_char *)message,strlen(message),sha1);
    		int qval = nsrl_bloom_query(b, sha1);	
    		if(!quiet){
	    		hex2string(nmessage, 40, sha1, 20);
	    		nmessage[40] = '\0';
	    		printf("%s sha1: %s is %s\n",message,nmessage,qval ? "present" : "absent");
			}
			if(qval) found++;
			string_increment(message);
    	}
    }
    else{
    	for(int i=0; i<count; i++){
		unsigned char hashbuf[20];
    	genprime(i,hashbuf);
		if(opt_x) continue;		/* don't lookup in the bloom filter */
		int qval = nsrl_bloom_query(b, hashbuf);
		if(!quiet){
	    	hex2string(message, 40, hashbuf, 20);
	    	message[40] = '\0';
	    	printf("%s is %s\n",message,qval ? "present" : "absent");
		}
		if(qval) found++;
    	}
    }
    printf("%d out of %d found.  %f rate.\n", found, count, (100*(float)found/(float)count));
    printf("bloom filter: %s\n",bloomfile);
    printf("vector bytes: %d\n",b->vector_bytes);
    printf("# hash_bits: %d\n",b->hash_bits);
    printf("# k=%d (hash_functions)\n",b->k);
    printf("# n=%u (Number of elements in BF)\n",(unsigned)1<<(b->M));
    printf("# comment: %s\n",b->comment);
    printf("# added items: %d\n",b->added_items);
    printf("# kusage: ");
    for(u_int i=0;i<b->k;i++){
	printf("%d ",nsrl_klookup_counters[i]);
    }
    printf("\n");

    /* Print usage information */
#ifdef HAVE_GETRUSAGE
    struct rusage ru;
    memset(&ru,0,sizeof(ru));
    if(getrusage(RUSAGE_SELF,&ru)==0){
	long total_msec =
	    ru.ru_utime.tv_sec*1000 + ru.ru_utime.tv_usec/1000 + 
	    ru.ru_stime.tv_sec*1000 + ru.ru_stime.tv_usec/1000;

	printf("# utime: %d.%03d  stime: %d.%03d\n",
	       (int)ru.ru_utime.tv_sec, (int)ru.ru_utime.tv_usec/1000,
	       (int)ru.ru_stime.tv_sec, (int)ru.ru_stime.tv_usec/1000);
	printf("# maxrss: %ld reclaims: %ld faults: %ld   swaps: %ld\n",
	       ru.ru_maxrss,ru.ru_minflt,ru.ru_majflt,ru.ru_nswap);
	printf("# inputs: %ld  outputs: %ld\n",ru.ru_inblock,ru.ru_oublock);
	printf("# total time: %ld msec\n",total_msec);
	printf("# average lookups per second: %.0f\n",count * 1000.0 / total_msec);
    }
#endif
    return 0;
}
